@@ -1,27 +1,40 @@
 use crate::conv_req::convert_req;
-use crate::Options;
 use actix_web::{http::StatusCode, web, HttpRequest, HttpResponse};
+use fmterr::fmt_err;
 use perseus::{
-    err_to_status_code, serve::get_page_for_template, ConfigManager, TranslationsManager,
+    errors::err_to_status_code,
+    internal::{
+        i18n::TranslationsManager,
+        serve::{get_page_for_template, GetPageProps, ServerOptions},
+    },
+    stores::{ImmutableStore, MutableStore},
 };
 use serde::Deserialize;
+use std::rc::Rc;
 
 #[derive(Deserialize)]
 pub struct PageDataReq {
     pub template_name: String,
+    pub was_incremental_match: bool,
 }
 
 /// The handler for calls to `.perseus/page/*`. This will manage returning errors and the like.
-pub async fn page_data<C: ConfigManager, T: TranslationsManager>(
+#[allow(clippy::too_many_arguments)]
+pub async fn page_data<M: MutableStore, T: TranslationsManager>(
     req: HttpRequest,
-    opts: web::Data<Options>,
-    config_manager: web::Data<C>,
+    opts: web::Data<Rc<ServerOptions>>,
+    immutable_store: web::Data<ImmutableStore>,
+    mutable_store: web::Data<M>,
     translations_manager: web::Data<T>,
+    global_state: web::Data<Option<String>>,
     web::Query(query_params): web::Query<PageDataReq>,
 ) -> HttpResponse {
     let templates = &opts.templates_map;
     let locale = req.match_info().query("locale");
-    let template_name = query_params.template_name;
+    let PageDataReq {
+        template_name,
+        was_incremental_match,
+    } = query_params;
     // Check if the locale is supported
     if opts.locales.is_supported(locale) {
         let path = req.match_info().query("filename");
@@ -31,8 +44,7 @@ pub async fn page_data<C: ConfigManager, T: TranslationsManager>(
             Ok(http_req) => http_req,
             // If this fails, the client request is malformed, so it's a 400
             Err(err) => {
-                return HttpResponse::build(StatusCode::from_u16(400).unwrap())
-                    .body(err.to_string())
+                return HttpResponse::build(StatusCode::from_u16(400).unwrap()).body(fmt_err(&err))
             }
         };
         // Get the template to use
@@ -45,12 +57,17 @@ pub async fn page_data<C: ConfigManager, T: TranslationsManager>(
             }
         };
         let page_data = get_page_for_template(
-            path,
-            locale,
+            GetPageProps {
+                raw_path: path,
+                locale,
+                was_incremental_match,
+                req: http_req,
+                global_state: &global_state,
+                immutable_store: immutable_store.get_ref(),
+                mutable_store: mutable_store.get_ref(),
+                translations_manager: translations_manager.get_ref(),
+            },
             template,
-            http_req,
-            config_manager.get_ref(),
-            translations_manager.get_ref(),
         )
         .await;
         match page_data {
@@ -59,7 +76,7 @@ pub async fn page_data<C: ConfigManager, T: TranslationsManager>(
                 http_res.content_type("text/html");
                 // Generate and add HTTP headers
                 for (key, val) in template.get_headers(page_data.state.clone()) {
-                    http_res.set_header(key.unwrap(), val);
+                    http_res.insert_header((key.unwrap(), val));
                 }
 
                 http_res.body(serde_json::to_string(&page_data).unwrap())
@@ -67,7 +84,7 @@ pub async fn page_data<C: ConfigManager, T: TranslationsManager>(
             // We parse the error to return an appropriate status code
             Err(err) => {
                 HttpResponse::build(StatusCode::from_u16(err_to_status_code(&err)).unwrap())
-                    .body(err.to_string())
+                    .body(fmt_err(&err))
             }
         }
     } else {
